/*
 *  Xfcalc version 0.1.6 - A multi-base calculator for X
 *  Copyright (C) 1997 Christopher John Purnell
 *                     cjp@sylph.u-net.com
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Toggle.h>

#include <math.h>
#include <errno.h>
#include <stdio.h>

#include "calc.h"

char bra_str[3];
char num_str[19];

double value = 0.0;
double mem_val = 0.0;
double drg = M_PI/180.0;
double base = 10.0;

int edit = 0;
int braket = 0;
int tok_ptr, num_ptr = 0;
char last = '0';
char tok_stack[43];
double num_stack[32] = { 0.0 };

char digit[16] = "0123456789ABCDEF";

void set_value(double v)
{
    value = v;
    last = 1;
    redraw();
}

void redraw(void)
{
    char *str = bra_str;
    char *ptr;
    double num, exp;
    int i, j, n, e;

    if(braket)
    {
	*(str++) = '(';
	*(str++) = digit[braket];
    }
    *str = 0;
    str = num_str;

    if(isnan(num = value))
    {
	strcpy(str, "NaN");
    }
    else
    {
	if((num ) < 0.0)
	{
	    num = -num;
	    *(str++) = '-';
	}

	if(isinf(num))
	{
	    strcpy(str, "Inf");
	}
	else
	{
	    double tmp = num;

	    e = (int)(exp = num ? floor(log(num) / log(base)) : 0.0);

	    if(e > 0 || e <= -4)
	    {
		num /= pow(base, exp);
	    }

	    if((num += pow(base, -11) / 2) >= base)
	    {
		num = tmp;
		e += 1;
		exp += 1.0;

		if(e > 0 || e <= -4)
		{
		    num /= pow(base, exp);
		}

		num += pow(base, -11) / 2;
	    }

	    if(e < 12 && e > -4)
		exp = 0.0;

	    if(e >= 12 || e < 0)
		e = 0;

	    ptr = str + 1;
	    if(!exp) ptr += e;

	    for(i = 0; i < 12; i++)
	    {
		*(str++) = digit[n = (int)(num)];

		if(n && str > ptr)
		    ptr = str;

		num -= (double)n; num *= base;

		if(i == e)
		    *(str++) = '.';
	    }

	    if(exp)
	    {
		*(ptr++) = exp < 0 ? exp = -exp, '-' : '+';

		exp += 0.5;

		for(j = 1; exp >= base; j++)
		    exp /= base;

		for(i = 0; i < j; i++)
		{
		    *(ptr++) = digit[n = (int)(exp)];
		    exp -= (double)n; exp *= base;
		}
	    }

	    *ptr = 0;
	}
    }

    refresh();
}

void refresh(void)
{
    Arg arg;

    XtSetArg(arg, XtNlabel, bra_str);
    XtSetValues(dis_opn, &arg, 1);
    arg.value = (XtArgVal)num_str;
    XtSetValues(dis_num, &arg, 1);
}

int get_modifiers(void)
{
    Arg arg;
    int i = 0, h = 0;

    XtSetArg(arg, XtNstate, &i);
    XtGetValues(inv, &arg, 1);
    arg.value = (XtArgVal)0;
    XtSetValues(inv, &arg, 1);

    arg.value = (XtArgVal)&h;
    XtGetValues(hyp, &arg, 1);
    arg.value = (XtArgVal)0;
    XtSetValues(hyp, &arg, 1);

    if(i) i = 1;
    if(h) i |= 2;

    return i;
}

void start_edit(char c)
{
    edit = ED_NUM;
    edit_len = (c == '0') ? 1 : 2;

    num_str[0] = ' ';
    num_str[1] = c;
    num_str[2] = 0;
}

void stop_edit(void)
{
    if(edit)
    {
	char c, a, *str = num_str;
	double v = 0.0, e;

	last = 1;
	edit = 0;

	while((c = *(++str)))
	{
	    if(c == '.' || c == '+' || c == '-')
		break;

	    c -= '0';
	    if(c > 9) c -= 'A'-('0'+10);
	    v *= base;
	    v += (double)(c);
	}

	if(c)
	{
	    if(c == '.')
	    {
		e = 1.0;

		while((c = *(++str)))
		{
		    if(c == '.' || c == '+' || c == '-')
			break;

		    c -= '0';
		    if(c > 9) c -= 'A'-('0'+10);
		    v *= base; e *= base;
		    v += (double)(c);
		    
		}

		v /= e;
	    }

	    if(c)
	    {
		e = 0.0;

		while((a = *(++str)))
		{
		    a -= '0';
		    if(a > 9) a -= 'A'-('0'+10);
		    e *= base;
		    e += (double)(a);
		}

		if(c == '-')
		    e = -e;

		v *= pow(base, e);
	    }
	}

	if(num_str[0] == '-')
	    v = -v;

	value = v;
    }
}
